<?php

/* ----------------------------------------------------------------------------
 *               INAF - National Institute for Astrophysics
 *               IRA  - Radioastronomical Institute - Bologna
 *               OATS - Astronomical Observatory - Trieste
 * ----------------------------------------------------------------------------
 *
 * Copyright (C) 2016 Istituto Nazionale di Astrofisica
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License Version 3 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

namespace RAP;

/**
 * MySQL implementation of the DAO interface. See comments on the DAO interface.
 */
class MySQLUserDAO extends BaseMySQLDAO implements UserDAO {

    public function __construct(Locator $locator) {
        parent::__construct($locator);
    }

    public function insertIdentity(Identity $identity, $userId) {

        $dbh = $this->getDBHandler();

        $stmt = $dbh->prepare("INSERT INTO identity(`user_id`, `type`, `email`, `name`, `surname`, `institution`, `typed_id`, `eppn`, `last_login`)"
                . " VALUES(:user_id, :type, :email, :name, :surname, :institution, :typed_id, :eppn, NOW())");

        $stmt->bindParam(':user_id', $userId);
        $stmt->bindParam(':type', $identity->type);
        $stmt->bindParam(':email', $identity->email);
        $stmt->bindParam(':name', $identity->name);
        $stmt->bindParam(':surname', $identity->surname);
        $stmt->bindParam(':institution', $identity->institution);
        $stmt->bindParam(':typed_id', $identity->typedId);
        $stmt->bindParam(':eppn', $identity->eppn);

        $stmt->execute();

        return $dbh->lastInsertId();
    }

    public function createUser() {

        $dbh = $this->getDBHandler();

        $stmt = $dbh->prepare("INSERT INTO user() VALUES()");
        $stmt->execute();

        return $dbh->lastInsertId();
    }

    private function getIdentityByRow($row) {

        $identity = new Identity($row['type']);

        $identity->id = $row['id'];
        $identity->primary = boolval($row['primary']);
        $identity->typedId = $row['typed_id'];
        $identity->email = $row['email'];
        $identity->name = $row['name'];
        $identity->surname = $row['surname'];
        $identity->institution = $row['institution'];
        $identity->eppn = $row['eppn'];

        return $identity;
    }

    public function findUserById(string $userId): ?User {

        if (!filter_var($userId, FILTER_VALIDATE_INT)) {
            return null;
        }

        $dbh = $this->getDBHandler();

        $stmt = $dbh->prepare("SELECT (u.`primary_identity` = i.`id`) AS `primary`,"
                . " i.`id`, `type`, `typed_id`, `email`, `name`, `surname`, `institution`, `eppn`"
                . " FROM identity i"
                . " JOIN `user` u on u.id = i.user_id"
                . " WHERE i.user_id = :user_id");

        $stmt->bindParam(':user_id', $userId);
        $stmt->execute();

        $result = $stmt->fetchAll();
        if (count($result) === 0) {
            return null;
        }

        $user = new User();
        $user->id = $userId;

        foreach ($result as $row) {
            $identity = $this->getIdentityByRow($row);
            $user->addIdentity($identity);
        }

        return $user;
    }

    public function setPrimaryIdentity($userId, $identityId) {

        $dbh = $this->getDBHandler();

        $stmt = $dbh->prepare("UPDATE `user` SET `primary_identity` = :identity_id WHERE `id` = :user_id");
        $stmt->bindParam(':identity_id', $identityId);
        $stmt->bindParam(':user_id', $userId);
        $stmt->execute();
    }

    public function findUserByIdentity($type, $identifier): ?User {

        $dbh = $this->getDBHandler();

        $query = "SELECT user_id FROM identity WHERE type = :type AND typed_id = :typed_id";

        $stmt = $dbh->prepare($query);
        $stmt->bindParam(':type', $type);
        $stmt->bindParam(':typed_id', $identifier);

        $stmt->execute();

        $result = $stmt->fetchAll();

        if (count($result) === 0) {
            return null;
        }
        if (count($result) > 1) {
            throw new Exception("Found multiple users associated to the same identity!");
        }

        $userId = $result[0]['user_id'];
        return $this->findUserById($userId);
    }

    public function searchUser($searchText) {

        $dbh = $this->getDBHandler();

        $query = "SELECT `user_id`, (u.`primary_identity` = i.`id`) AS `primary`,"
                . " i.`id`, `type`, `typed_id`, `email`, `name`, `surname`, `institution`, `eppn`"
                . " FROM identity i"
                . " JOIN `user` u on u.id = i.user_id"
                . " WHERE i.user_id IN"
                . " (SELECT user_id FROM identity"
                . " WHERE `email` LIKE :email OR `email` LIKE :emailPart"
                . " OR `eppn` LIKE :eppn"
                . " OR `name` LIKE :name OR `surname` LIKE :surname"
                . " OR CONCAT(`name`,' ',`surname`) LIKE :namesurname)";

        $stmt = $dbh->prepare($query);

        $searchParam = $searchText . '%';
        if (count_chars($searchText) > 4) {
            $searchParam = '%' . $searchParam;
        }
        $emailPartSearchParam = '%.' . $searchText . '%';
        $stmt->bindParam(':email', $searchParam);
        $stmt->bindParam(':emailPart', $emailPartSearchParam);
        $stmt->bindParam(':eppn', $searchParam);
        $stmt->bindParam(':name', $searchParam);
        $stmt->bindParam(':surname', $searchParam);
        $stmt->bindParam(':namesurname', $searchParam);

        return $this->getUsersListFromStatement($stmt);
    }

    public function getUsers(array $identifiers): array {

        if (count($identifiers) === 0) {
            return [];
        }

        $dbh = $this->getDBHandler();

        $query = "SELECT `user_id`, (u.`primary_identity` = i.`id`) AS `primary`,"
                . " i.`id`, `type`, `typed_id`, `email`, `name`, `surname`, `institution`, `eppn`"
                . " FROM identity i"
                . " JOIN `user` u on u.id = i.user_id"
                . " WHERE i.user_id IN (";

        $first = true;
        foreach ($identifiers as $id) {
            if (!$first) {
                $query .= ',';
            }
            $query .= ':id_' . $id;
            if ($first) {
                $first = !$first;
            }
        }

        $query .= ')';

        $stmt = $dbh->prepare($query);

        foreach ($identifiers as &$id) {
            $stmt->bindParam(':id_' . $id, $id);
        }

        return $this->getUsersListFromStatement($stmt);
    }

    private function getUsersListFromStatement(\PDOStatement $stmt): array {

        $stmt->execute();

        $userMap = array();

        foreach ($stmt->fetchAll() as $row) {

            $identity = $this->getIdentityByRow($row);

            $userId = $row['user_id'];
            if (array_key_exists($userId, $userMap)) {
                $user = $userMap[$userId];
            } else {
                $user = new User();
                $user->id = $userId;
                $userMap[$userId] = $user;
            }

            array_push($user->identities, $identity);
        }

        $users = [];
        foreach ($userMap as $userId => $user) {
            array_push($users, $user);
        }
        return $users;
    }

    public function joinUsers($userId1, $userId2) {
        $dbh = $this->getDBHandler();

        try {
            $dbh->beginTransaction();

            // Deleting keep_separated pairs
            $stmt1 = $dbh->prepare("DELETE FROM keep_separated WHERE "
                    . "(user_id1 = :id1 AND user_id2 = :id2) OR"
                    . "(user_id1 = :id2 AND user_id2 = :id1)");
            $stmt1->bindParam(':id1', $userId1);
            $stmt1->bindParam(':id2', $userId2);
            $stmt1->execute();
            
            // Deleting keep_separated records of user that is going to be deleted
            $stmt2 = $dbh->prepare("DELETE FROM keep_separated WHERE user_id1 = :id2 OR user_id2 = :id2");
            $stmt2->bindParam(':id2', $userId2);
            $stmt2->execute();
            
            // Moving identities from user2 to user1
            $stmt3 = $dbh->prepare("UPDATE `identity` SET `user_id` = :id1 WHERE `user_id` = :id2");
            $stmt3->bindParam(':id1', $userId1);
            $stmt3->bindParam(':id2', $userId2);
            $stmt3->execute();

            // Deleting user2
            $stmt4 = $dbh->prepare("DELETE FROM `user` WHERE `id` = :id2");
            $stmt4->bindParam(':id2', $userId2);
            $stmt4->execute();

            $dbh->commit();
        } catch (Exception $ex) {
            $dbh->rollBack();
            throw $ex;
        }
    }

    function isAdmin($userId): bool {

        $dbh = $this->getDBHandler();

        $query = "SELECT user_id FROM rap_permissions WHERE permission = 'ADMIN' AND user_id = :userId";

        $stmt = $dbh->prepare($query);
        $stmt->bindParam(':userId', $userId);
        $stmt->execute();

        $result = $stmt->fetchAll();

        return count($result) === 1;
    }

    function updateIdentity(Identity $identity): void {

        $dbh = $this->getDBHandler();

        $query = "UPDATE identity SET email = :email, name = :name, surname = :surname, institution = :institution,"
                . " last_login = NOW() WHERE id = :id";

        $stmt = $dbh->prepare($query);
        $stmt->bindParam(':email', $identity->email);
        $stmt->bindParam(':name', $identity->name);
        $stmt->bindParam(':surname', $identity->surname);
        $stmt->bindParam(':institution', $identity->institution);
        $stmt->bindParam(':id', $identity->id);

        $stmt->execute();
    }

    function findJoinableUsersByEmail(string $email): array {

        $dbh = $this->getDBHandler();

        $query = "SELECT DISTINCT(i3.user_id)\n"
                . "FROM user u\n"
                . "JOIN identity i ON i.user_id = u.id\n"
                . "JOIN identity i2 ON i.user_id = i2.user_id\n"
                . "JOIN identity i3 ON i2.email = i3.email\n"
                . "WHERE i.email = :email";

        $stmt = $dbh->prepare($query);
        $stmt->bindParam(':email', $email);

        $stmt->execute();

        $results = [];
        foreach ($stmt->fetchAll() as $row) {
            array_push($results, $row['user_id']);
        }
        return $results;
    }

    function findJoinableUsersByUserId(string $userId): array {

        $dbh = $this->getDBHandler();

        $query = "SELECT DISTINCT(i3.user_id)\n"
                . "FROM user u\n"
                . "JOIN identity i ON i.user_id = u.id\n"
                . "JOIN identity i2 ON i.user_id = i2.user_id\n"
                . "JOIN identity i3 ON i2.email = i3.email\n"
                . "WHERE i.email IN (SELECT email FROM identity where user_id = :user_id)\n"
                . "AND i3.user_id NOT IN (SELECT :user_id UNION\n"
                . "SELECT user_id2 FROM keep_separated WHERE user_id1 = :user_id\n"
                . "UNION\n"
                . "SELECT user_id1 FROM keep_separated WHERE user_id2 = :user_id)";

        $stmt = $dbh->prepare($query);
        $stmt->bindParam(':user_id', $userId);

        $stmt->execute();

        $results = [];
        foreach ($stmt->fetchAll() as $row) {
            array_push($results, $row['user_id']);
        }
        return $results;
    }

    function insertRejectedJoin(string $userId1, string $userId2): void {

        $dbh = $this->getDBHandler();

        $query = "INSERT INTO keep_separated (user_id1, user_id2) VALUES (:id1, :id2)\n"
                . "ON DUPLICATE KEY UPDATE user_id1 = user_id1, user_id2 = user_id2";

        $stmt = $dbh->prepare($query);
        $stmt->bindParam(':id1', $userId1);
        $stmt->bindParam(':id2', $userId2);

        $stmt->execute();
    }

}
